/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace SudokuSolver
{
    /// <summary>
    /// This struct is used to supply the spwaned threads with the 
    /// information they need to do the calculations.
    /// </summary>
    struct SudokuFieldAndHandle
    {
        /// <summary>
        /// The SudokuCellField the thread should work on.
        /// </summary>
        public SudokuCellField m_currentField;

        /// <summary>
        /// The ManualResetEvent that the thread should trigger when it
        /// is done with the calculations.
        /// </summary>
        public ManualResetEvent m_resetEvent;
    }

    /// <summary>
    /// This class encapsulates the methods to solve a complete Sudoku game. It 
    /// holds the state of the Sudoku using SudokuCellFields. To improve the performance
    /// on multi-core systems, it spawns new threads in the calculation process, when it
    /// encounters a Sudokulayout where no logic based conclusions can be found anymore. 
    /// It then uses a "brute-force" guess algorithm.
    /// </summary>
    class Sudoku
    {
        /// <summary>
        /// This Delegate is used as a callback when this class is finished with
        /// the calculations on the current Sudoku.
        /// </summary>
        public delegate void resultsFound();

        /// <summary>
        /// This Delegate is used to update the progress bar of the solving animation 
        /// window by means of a callback function.
        /// </summary>
        /// <param name="progress">This parameter represents the precentage of the progress bar. It should be between 0 and 100.</param>
        public delegate void updateProgress(int progress);

        // variables used to save the state of the Sudoku

        /// <summary>
        /// Array of the solutions found (or the current working layout).
        /// </summary>
        private ArrayList m_cellFields;

        /// <summary>
        /// Holds the number of the current shown Sudoku (in the case there are several solutions)
        /// </summary>
        private int m_currentShownSolution = 0;

        /// <summary>
        /// Holds the information if there exists a complete solution for this Sudoku.
        /// </summary>
        private bool m_completeSolutionExists = false;

        /// <summary>
        /// Holds the delegate that is used as a callback method to update the progress bar in
        /// the solving animation window.
        /// </summary>
        private updateProgress m_progressMethod;

        /// <summary>
        /// Holds the information if the user wants to cancel the solving progress.
        /// </summary>
        private bool m_cancelCalculation = false;
        
        /// <summary>
        /// Standard constructor that fills the solution array with an empty SudokuCellField.
        /// </summary>
        public Sudoku()
        {
            // constructor
            m_cellFields = new ArrayList(1);
            m_cellFields.Add(new SudokuCellField());
        }

        /// <summary>
        /// Copy constructor that copies the supplied Sudoku class.
        /// </summary>
        /// <param name="copy">The instance that should be copied.</param>
        public Sudoku(Sudoku copy)
        {
            // copy all SudokuCellFields
            m_cellFields = new ArrayList(copy.m_cellFields.Count);
            for (int i = 0; i < copy.m_cellFields.Count; i++)
                m_cellFields.Add(new SudokuCellField((SudokuCellField)copy.m_cellFields[i]));

            // copy all other variables
            m_currentShownSolution = copy.m_currentShownSolution;
            m_completeSolutionExists = copy.m_completeSolutionExists;
            m_progressMethod = copy.m_progressMethod;
            m_cancelCalculation = copy.m_cancelCalculation;
        }

        /// <summary>
        /// Set the given number on the current active SudokuCellField. This fixes the supplied
        /// number at the indicated cell. A fixed number excludes this number at the row of the
        /// cell, at the column of the cell and at the panel of the cell.
        /// </summary>
        /// <param name="j">Row of the cell that should be fixed. Valid values are between 0 and 8.</param>
        /// <param name="i">Column of the cell that should be fixed. Valid values are between 0 and 8.</param>
        /// <param name="number">The number the cell should be fixed to. Valid values are between 0 and 8.</param>
        public void fixNumber(int j, int i, int number)
        {
            // check if i and j are with the boundaries
            if ((i < 0) || (i > 8)) return;
            if ((j < 0) || (j > 8)) return;
            if ((number < 0) || (number > 8)) return;

            // fixes the given number within the row, column and panel
            ((SudokuCellField)m_cellFields[m_currentShownSolution]).fixNumber(j, i, number);
        }

        /// <summary>
        /// Returns if the indicated cell of the current SudokuCellField has a fixed number. In the
        /// case that a fixed number exists, the number is returned in the index parameter. If parameter
        /// are out of bounds this method returns false.
        /// </summary>
        /// <param name="j">Row of the cell that should be read. Valid values are between 0 and 8.</param>
        /// <param name="i">Column of the cell that should be read. Valid values are between 0 and 8.</param>
        /// <param name="index">The number the cell should be read. This variable is set to the number of the cell,
        /// if the cell has a fixed number.</param>
        /// <returns>True if the indicated cell has a fixed number, False otherwise.</returns>
        public bool getNumber(int j, int i, ref int index)
        {
            // check if i and j are with the boundaries
            if ((i < 0) || (i > 8)) return false;
            if ((j < 0) || (j > 8)) return false;

            return ((SudokuCellField)m_cellFields[m_currentShownSolution]).getNumber(j, i, ref index);
        }

        /// <summary>
        /// This thread-safe method is called by the worker threads if they find a
        /// solution. When this method is called, it checks if the solution provided
        /// is new, or if it is already in the solution array.
        /// </summary>
        /// <param name="field">The Sudoku field which holds the solution to the Sudoku.</param>
        private void addSolution(ref SudokuCellField field)
        {
            // lock on the array to make this method thread-safe
            lock (m_cellFields)
            {
                // check if the found solutions are truely new solutions
                bool truelyNew = true;
                for (int i = 0; i < m_cellFields.Count; i++)
                {
                    // check against new field
                    if (((SudokuCellField)m_cellFields[i]) == field)
                    {
                        truelyNew = false;
                        break;
                    }
                }
                if (truelyNew) m_cellFields.Add(field);
            }
        }

        /// <summary>
        /// This method is needed because only one parameter can be passed to threads. It
        /// is called by the ThreadPool. The stateInfo parameter is a struct that holds the 
        /// two parameters needed: The Sudoku field the thread should work on and the 
        /// ManualResetEvent to indicate, that a thread is finished. When the thread has done
        /// its work, it sets the ManualResetEvent.
        /// </summary>
        /// <param name="stateInfo">The struct that holds the Sudoku field and the ManualResetEvent.</param>
        private void recursiveSolveThread(Object stateInfo)
        {
            // get event handle and the SudokuCellField
            SudokuFieldAndHandle fieldAndHandle = (SudokuFieldAndHandle)stateInfo;
            SudokuCellField currentField = fieldAndHandle.m_currentField;
            recursiveSolve(ref currentField, false);

            // signal calling thread that this thread is finished
            fieldAndHandle.m_resetEvent.Set();
        }

        /// <summary>
        /// This is the main method that implements the algorithm. It is a recursive
        /// method which gets called by the worker threads as well as the main calculation
        /// thread. To improve performance, worker threads are not allowed to spawn threads
        /// themselves.
        /// </summary>
        /// <param name="currentField">The current Sudoku field this method works on.</param>
        /// <param name="spawnNewThreads">Indicates if this thread may spawn new worker threads.</param>
        private void recursiveSolve(ref SudokuCellField currentField, bool spawnNewThreads)
        {
            // did we find a solution?
            bool result = true;

            // still more work to do?
            bool stillNotDone = true;

            while (stillNotDone && (!m_cancelCalculation))
            {
                int fixedPositionsFound = 0;

                // check if this field is still valid after 
                // a new possible fixed number of the recursive descent
                if (!(currentField.isValid())) return;

                // check for naked pairs
                currentField.checkForNakedPairs();

                // finds all fields that have only one possibility left
                fixedPositionsFound += currentField.findFixedPositions();

                // finds out for each number if there is only one position left in each column, row and panel
                fixedPositionsFound += currentField.checkNumberPositions();

                // use brute force to set certain fields to a possible value to find a solution
                if (fixedPositionsFound == 0)
                {
                    // list of all wait handles that the main thread should wait for
                    ArrayList waitHandles = new ArrayList();

                    // find possible position where several numbers can be used
                    for (int j = 0; j < 9; j++)
                    {
                        for (int i = 0; i < 9; i++)
                        {
                            if (currentField.needsProcessing(j, i))
                            {
                                // use a recursive depth-first-search to check for all possibilities
                                // check all numbers
                                for (int k = 0; k < 9; k++)
                                {
                                    // if this number is a possibility
                                    if (currentField.isPossible(j, i, k))
                                    {
                                        // start new recursive approach with a copy of the current field
                                        SudokuCellField newField = new SudokuCellField(currentField);

                                        // fix this potential number
                                        newField.fixNumber(j, i, k);

                                        // do recursive descent todo
                                        if (spawnNewThreads)
                                        {
                                            // create new wait handle and fill struct
                                            SudokuFieldAndHandle fieldAndHandle;
                                            fieldAndHandle.m_currentField = newField;
                                            fieldAndHandle.m_resetEvent = new ManualResetEvent(false);

                                            // enqueue new thread in threadpool 
                                            ThreadPool.QueueUserWorkItem(new WaitCallback(recursiveSolveThread), fieldAndHandle);

                                            // add wait handle to the array
                                            waitHandles.Add(fieldAndHandle.m_resetEvent);
                                        }
                                        else
                                        {
                                            recursiveSolve(ref newField, false);
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if ((spawnNewThreads) && (waitHandles.Count > 0))
                    {
                        // because the WaitAll() method can only wait on 64
                        // handles at a time we have to wait one by one on every handle

                        int signaledEvents = 0;
                        while ((signaledEvents < waitHandles.Count) && (!m_cancelCalculation))
                        {
                            // reset signaled events
                            signaledEvents = 0;

                            // wait on every single handle that is in the handle list
                            // do some kind of half busy wait 
                            for (int i = 0; i < waitHandles.Count; i++)
                            {
                                // wait on handle
                                ManualResetEvent eventHandle = (ManualResetEvent)waitHandles[i];
                                if (eventHandle.WaitOne(1, false)) signaledEvents++;

                                // exit for loop if the user wants the threads to exit
                                if (m_cancelCalculation) break;
                            }

                            // update progress in GUI
                            Double progress = Convert.ToDouble(signaledEvents) / Convert.ToDouble(waitHandles.Count);
                            int progressInt = Convert.ToInt32(progress * 100.0);
                            if (!m_cancelCalculation)
                            {
                                try
                                {
                                    m_progressMethod(progressInt);
                                }
                                catch (ObjectDisposedException)
                                {
                                    // ignore... this exception only occus if the user cancels 
                                    // the calculations and in this case all threads close down
                                    // anyway
                                }
                            }

                            // put the thread to sleep to reduce the load of the half busy wait
                            if (!m_cancelCalculation) Thread.Sleep(200);
                        }
                    }
                }

                // check for cells still to solve
                bool foundOne = currentField.stillOpenCells();

                // emergency break when nothing is improving
                if ((fixedPositionsFound == 0) && (foundOne))
                {
                    foundOne = false;
                    result = false;
                }

                stillNotDone = foundOne;
            }

            // if a solution was found, add it to the solution array
            if (result) addSolution(ref currentField);
        }

        /// <summary>
        /// This method is needed to spawn the main calculation thread. It starts
        /// the recursive solving algorithm and notifies the caller when the
        /// calculations are done. Timing methods are implemented for debugging
        /// purposes.
        /// </summary>
        /// <param name="data">The callback method that is used to notify the caller.</param>
        private void startSolving(object data)
        {
            // cast the data back to the callback 
            resultsFound callback = (resultsFound)data;

            //// timing for the debug message
            //// take the start time
            //DateTime beginning = DateTime.Now;

            // define the current Sudoku field
            SudokuCellField currentField = ((SudokuCellField)m_cellFields[0]);

            // start the recursive solve algorithm
            recursiveSolve(ref currentField, true);

            //// timing for the debug message
            //// calculate elapsed time
            //DateTime end = DateTime.Now;
            //TimeSpan duration = end - beginning;

            // if the calculation was not canceled
            if (!m_cancelCalculation)
            {

                // check if a complete solution exists
                if (m_cellFields.Count == 1)
                {
                    m_completeSolutionExists = !((SudokuCellField)m_cellFields[0]).stillOpenCells();
                }
                // delete partial solution if several solution were found
                else if (m_cellFields.Count > 1)
                {
                    if (((SudokuCellField)m_cellFields[0]).stillOpenCells()) m_cellFields.RemoveAt(0);
                    m_completeSolutionExists = true;
                }

                //// timing for the debug message
                //MessageBox.Show("Duration of solving the Sudoku = " + duration.ToString());

                // tell the caller that solution(s) were found
                callback();
            }
        }

        /// <summary>
        /// This method gets called by the outside class that wants this Sudoku solved. It must
        /// supply two delegate methods that are used as callbacks. The first one will be
        /// called when the calculation is finish. The second one will be called in regular
        /// time slots to update a progress counter.
        /// </summary>
        /// <param name="callback">The callback method that is called when the calculation is done.</param>
        /// <param name="progress">The update method that is called to update the progress counter.</param>
        public void solve(resultsFound callback, updateProgress progress)
        {
            // start new thread to decouple GUI from solving the Sudoku
            Thread solvingThread = new Thread(new ParameterizedThreadStart(startSolving));

            // set method for progress update
            m_progressMethod = progress;

            // start thread
            solvingThread.Start(callback);
        }

        /// <summary>
        /// This method returns the number of solutions that are saved in this
        /// Sudoku class. A partial solution will be counted as solution as well.
        /// </summary>
        /// <returns>The number of solutions</returns>
        public int getNumberOfSolutions()
        {
            lock (m_cellFields)
            {
                return m_cellFields.Count;
            }
        }

        /// <summary>
        /// This method returns if a complete solution for this Sudoku exists.
        /// </summary>
        /// <returns>True if a complete solution exists, False otherwise.</returns>
        public bool completeSolutionExists()
        {
            return m_completeSolutionExists;
        }

        /// <summary>
        /// This method switches the current solution to the next solution if
        /// there is another one. To make it easier for the user to distinguish
        /// the solutions, no wrap around is supported.
        /// </summary>
        public void showNextSolution()
        {
            // switch the current solution to the next solution
            if (m_cellFields.Count > 1)
            {
                m_currentShownSolution++;
                if (m_currentShownSolution > (m_cellFields.Count - 1))
                {
                    m_currentShownSolution = m_cellFields.Count - 1;
                }
            }
        }

        /// <summary>
        /// This method switches the current solution to the previous solution if
        /// there is another one. To make it easier for the user to distinguish
        /// the solutions, no wrap around is supported.
        /// </summary>
        public void showPreviousSolution()
        {
            // switch the current solution to the previous solution
            if (m_cellFields.Count > 1)
            {
                m_currentShownSolution--;
                if (m_currentShownSolution < 0)
                {
                    m_currentShownSolution = 0;
                }
            }
        }

        /// <summary>
        /// This method should be called by the GUI if the user wants to cancel
        /// the calculation of the Sudoku.
        /// </summary>
        public void cancelCalculation()
        {
            // not thread safe, but ok because
            // this value can only be changed to true once
            // the threads will notice the new value after the 
            // next evaluation
            m_cancelCalculation = true;
        }
    }
}
