/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;

namespace SudokuSolver
{
    /// <summary>
    /// This class represents one cell on the board of a Sudoku game.
    /// For performance reasons it uses bool variables to save for each
    /// number if this number is still possible in this cell.
    /// </summary>
    class SudokuCell
    {
        /// <summary>
        /// Holds the information if this cell needs further processing or
        /// if it has already a fixed number. This variable is true in the 
        /// beginning and can only be changed to false when the number for
        /// this cell gets fixed.
        /// </summary>
        private bool m_needsProcessing = true;

        /// <summary>
        /// Array of bool variables that save for each number, if this number is
        /// still possible in this cell.
        /// </summary>
        private bool[] m_possibleNumbers = new bool[9];

        /// <summary>
        /// Variable to hold a potential fixed number. This information could be
        /// retrieved on the fly from the m_possibleNumbers array. To increase 
        /// performance it is saved in its own variable.
        /// </summary>
        private int m_fixedNumber = -1;

        /// <summary>
        /// Number of free possibilities that are left. This information could be
        /// retrieved on the fly from the m_possibleNumbers array. To increase 
        /// performance it is saved in its own variable.
        /// </summary>
        private int m_freeNumbers = 9;

        /// <summary>
        /// Standard constructor.
        /// </summary>
        public SudokuCell()
        {
            for (int i = 0; i < 9; i++) m_possibleNumbers[i] = true;
        }

        /// <summary>
        /// Copy constructor.
        /// </summary>
        /// <param name="copy">The SudokuCell that should be copied.</param>
        public SudokuCell(SudokuCell copy)
        {
            // copy all possibilities
            for (int i = 0; i < 9; i++) m_possibleNumbers[i] = copy.m_possibleNumbers[i];

            // copy all variables
            m_needsProcessing = copy.m_needsProcessing;
            m_fixedNumber = copy.m_fixedNumber;
            m_freeNumbers = copy.m_freeNumbers;
        }

        /// <summary>
        /// This method returns if this cell needs further processing or if
        /// it already has a fixed number. If this method retruns true, but
        /// freeNumbers() returns 0, the Sudoku can not be solved because of
        /// a logic error.
        /// </summary>
        /// <returns>True if it needs further processing, False if this cell has already a fixed number.</returns>
        public bool needsProcessing()
        {
            return m_needsProcessing;
        }

        /// <summary>
        /// This method returns the number that was fixed for this cell.
        /// If this cell has no fixed number yet, -1 will be returned.
        /// </summary>
        /// <returns>-1 if no number was fixed yet, between 0 and 8 otherwise.</returns>
        public int getFixedNumber()
        {
            return m_fixedNumber;
        }

        /// <summary>
        /// Fixes the supplied index for this cell. Because of performance 
        /// considerations, the index is not checked against boundaries.
        /// </summary>
        /// <param name="index">The index that should be fixed. Needs to be between 0 and 8.</param>
        public void fixNumber(int index)
        {
            m_fixedNumber = index;
            m_needsProcessing = false;
            m_freeNumbers = 1;
            for (int i = 0; i < 9; i++) m_possibleNumbers[i] = (i == index);
        }

        /// <summary>
        /// This method returns the number of possibilities left for this cell.
        /// </summary>
        /// <returns>Value between 0 and 9.</returns>
        public int freeNumbers()
        {
            return m_freeNumbers;
        }

        /// <summary>
        /// This method removes on possibility for this cell. The possibility removed
        /// is given by the index. Because of performance considerations, the index is
        /// not checked against boundaries.
        /// </summary>
        /// <param name="index">The index that is not possible for this cell. Needs to be between 0 and 8.</param>
        public void removePossibility(int index)
        {
            if (m_possibleNumbers[index]) m_freeNumbers--;
            m_possibleNumbers[index] = false;
        }

        /// <summary>
        /// This method returns if the given index is still possible for this cell. 
        /// Because of performance considerations, the index is not checked against 
        /// boundaries
        /// </summary>
        /// <param name="index">The index in question for this cell. Needs to be between 0 and 8.</param>
        /// <returns>True if the index is possible, False otherwise.</returns>
        public bool isPossible(int index)
        {            
            return m_possibleNumbers[index];
        }

        /// <summary>
        /// This method checks if there is only one possibility left for this cell.
        /// If so, it fixes this number and returns true.
        /// </summary>
        /// <returns>True if only one possibility was left and the number got fixed, False otherwise.</returns>
        public bool checkForOnePossibilityLeft() 
        {
            if (m_needsProcessing)
            {
                // checks if only one possibility is left
                if (m_freeNumbers == 1)
                {
                    // if only one possible number is left, find it
                    int index = -1;
                    for (int i = 0; i < 9; i++)
                    {
                        if (m_possibleNumbers[i])
                        {
                            index = i;
                            break;
                        }
                    }

                    // fix this number
                    m_fixedNumber = index;
                    m_needsProcessing = false;
                    for (int i = 0; i < 9; i++) m_possibleNumbers[i] = (i == index);
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// This method returns all remaining possibilities in one integer value.
        /// This allows to check if several cells have the same possibilities.
        /// </summary>
        /// <returns>An integer which is unique for each combination of possibilities.</returns>
        public int possiblities()
        {
            int result = 0;
            int n = 1;

            // give each possibility a digit in the result
            for (int i = 0; i < 9; i++)
            {
                if (m_possibleNumbers[i])
                {
                    result += ((i+1) * n);
                    n *= 10;
                }
            }

            return result;
        }

        /// <summary>
        /// This method checks if this cell is still in a valid state.
        /// </summary>
        /// <returns>True if this cell is still valid, False otherwise.</returns>
        public bool isValid()
        {
            if ((m_needsProcessing) && (m_freeNumbers >= 1)) return true;
            else if ((!m_needsProcessing) && (m_freeNumbers == 1)) return true;
            else return false;
        }

        /// <summary>
        /// This overloaded operator is implemented so that two cells can be compared.
        /// </summary>
        /// <param name="cell1">First cell to compare.</param>
        /// <param name="cell2">Second cell to compare.</param>
        /// <returns>True if the two cells are equal, False otherwise.</returns>
        public static bool operator ==(SudokuCell cell1, SudokuCell cell2)
        {
            if (cell1.m_needsProcessing != cell2.m_needsProcessing) return false;
            if (cell1.m_fixedNumber != cell2.m_fixedNumber) return false;
            if (cell1.m_freeNumbers != cell2.m_freeNumbers) return false;
            for (int i = 0; i < 9; i++)
                if (cell1.m_possibleNumbers[i] != cell2.m_possibleNumbers[i]) return false;
            return true;
        }

        /// <summary>
        /// This overloaded operator needs to be implemented as soon as the
        /// ==-operator is implemented.
        /// </summary>
        /// <param name="cell1">First cell to compare.</param>
        /// <param name="cell2">Second cell to compare.</param>
        /// <returns>True if the two cells are not equal, False otherwise.</returns>
        public static bool operator !=(SudokuCell cell1, SudokuCell cell2)
        {
            return !(cell1 == cell2);
        }

        /// <summary>
        /// This overloaded operator needs to be implemented as soon as the 
        /// ==-operator is implemented. It returns, if two objects are equal.
        /// </summary>
        /// <param name="obj">The other object that will be compared to this cell.</param>
        /// <returns>True if the two objects are equal, False otherwise.</returns>
        public override bool Equals(object obj)
        {
            if (obj == null) return false;

            if (this.GetType() != obj.GetType()) return false;

            // safe because of the GetType check
            SudokuCell cell = (SudokuCell)obj;

            // check members
            if (!Object.Equals(m_possibleNumbers, cell.m_possibleNumbers)) return false;
            if (!m_needsProcessing.Equals(cell.m_needsProcessing)) return false;
            if (!m_fixedNumber.Equals(cell.m_fixedNumber)) return false;
            if (!m_freeNumbers.Equals(cell.m_freeNumbers)) return false;

            return true;
        }

        /// <summary>
        /// This Method needs to be overriden to resolve a warning. It is not implemented
        /// and will raise a NotImplementedException if called anyway. This method will
        /// only be called if the cell class is used in a hashmap.
        /// </summary>
        /// <returns>Nothing, throws Exception if called.</returns>
        public override int GetHashCode()
        {
            throw new NotImplementedException();
        }
    }
}

