/************************************************************************
 * Copyright (c) 2008 by Gregor Ehrenstein (grx@pandacube.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.IO;
using System.Threading;

namespace SudokuSolver
{
    public partial class SudokuSolverForm1 : Form
    {
        // variables to save the state of the form
        private Sudoku m_currentSudoku;
        private Color m_fixedColor = Color.OrangeRed;
        private Color m_backgroundColor = Color.White;
        private Color m_solutionColor = Color.SpringGreen;
        private SolveAnimationForm1 m_animationForm;
        private AboutForm1 m_aboutForm = new AboutForm1();
        
        /// <summary>
        /// Standard constructor that initializes the GUI.
        /// </summary>
        public SudokuSolverForm1()
        {
            InitializeComponent();
        }

        /// <summary>
        /// This method fills the comboboxes with the current values in the m_currentSudoku.
        /// </summary>
        private void fillComboboxes()
        {
            int index = 0;
            // fill in the current solution
            for (int j = 0; j < 9; j++)
            {
                for (int i = 0; i < 9; i++)
                {
                    // only fill with proven values
                    if (m_currentSudoku.getNumber(j, i, ref index))
                    {
                        if (this.comboBoxes[j, i].SelectedIndex != index)
                        {
                            this.comboBoxes[j, i].SelectedIndex = index;
                            this.comboBoxes[j, i].BackColor = m_solutionColor;
                        }
                    }
                    else
                    {
                        this.comboBoxes[j, i].SelectedIndex = -1;
                        this.comboBoxes[j, i].BackColor = m_backgroundColor;
                    }
                }
            }
        }

        /// <summary>
        /// This method loads the sudoku file and fills the comboboxes of the GUI.
        /// If any error occurs, the interface gets reset and MessageBox with an
        /// error message will be displayed.
        /// </summary>
        /// <param name="filename">The path and name of the Sudoku file that should be loaded.</param>
        private void LoadFile(String filename)
        {
            try
            {
                // Create an instance of StreamReader to read from the file.
                // The using statement also closes the StreamReader.
                using (StreamReader sr = new StreamReader(filename))
                {
                    String line;
                    // Read 9 lines from the file 
                    for(int j = 0; j < 9; j++)
                    {
                        line = sr.ReadLine();
                        string[] singleValues = line.Split(' ');

                        // the strings need 9 values per line
                        if (singleValues.Length < 9) {
                            MessageBox.Show("The Format of the Sudoku file is incorrect. The file needs to be a space-seperated 9 by 9 field of integer values between 0 and 9.", "The file could not be read...");
                            ResetButton1_Click(this, new EventArgs());
                            return;
                        }

                        // load each value
                        for (int i = 0; i < 9; i++)
                        {
                            int index = Convert.ToInt32(singleValues[i]);
                            // check for boundaries
                            if ((index >= 0) && (index <= 9)) this.comboBoxes[j, i].SelectedIndex = index - 1;
                            else
                            {
                                MessageBox.Show("The Format of the Sudoku file is incorrect. The file needs to be a space-seperated 9 by 9 field of integer values between 0 and 9.", "The file could not be read...");
                                ResetButton1_Click(this, new EventArgs());
                                return;
                            }
                        }
                    }
                }
            }
            catch (Exception e)
            {
                // Let the user know what went wrong
                MessageBox.Show(e.Message, "The file could not be read...");
                ResetButton1_Click(this, new EventArgs());
            }
        }

        /// <summary>
        /// This methods saves the current values in the comboboxes to a Sudoku file.
        /// </summary>
        /// <param name="filename">The path and name of the Sudoku file that should be saved.</param>
        private void SaveFile(String filename)
        {
            try
            {
                using (StreamWriter sw = File.CreateText(filename))
                {
                    string line;
                    // write file, line by line
                    for (int j = 0; j < 9; j++)
                    {
                        line = "";
                        for (int i = 0; i < 9; i++)
                        {
                            int index = this.comboBoxes[j, i].SelectedIndex + 1;
                            line += Convert.ToString(index) + " ";
                        }
                        sw.WriteLine(line);
                    }
                    sw.Close();
                }
            }
            catch (Exception e)
            {
                // Let the user know what went wrong
                MessageBox.Show(e.Message, "The file could not be written...");
            }
        }

        /// <summary>
        /// This method gets called when the about button is clicked. It
        /// will display the about window.
        /// </summary>
        private void AboutButton1_Click(object sender, EventArgs e)
        {
            // show the about window
            m_aboutForm.ShowDialog();
        }

        /// <summary>
        /// This method gets called when the prev botton was clicked. It
        /// will display the previous solution in the case that solving
        /// the Sudoku showed, that the Sudoku has several solutions.
        /// </summary>
        private void PrevButton1_Click(object sender, EventArgs e)
        {
            // show the previous solution
            m_currentSudoku.showPreviousSolution();
            fillComboboxes();
        }

        /// <summary>
        /// This method gets called when the next botton was clicked. It
        /// will display the next solution in the case that solving
        /// the Sudoku showed, that the Sudoku has several solutions.
        /// </summary>
        private void NextButton1_Click(object sender, EventArgs e)
        {
            // show the next solution
            m_currentSudoku.showNextSolution();
            fillComboboxes();
        }

        /// <summary>
        /// This method gets called when the Load button was clicked. It
        /// will open a file dialog to enable the user to define the Sudoku
        /// file that he wants to be loaded.
        /// </summary>
        private void LoadButton1_Click(object sender, EventArgs e)
        {
            // open 'open file dialog'
            if (m_openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                // reset comboboxes first
                ResetButton1_Click(this, new EventArgs());

                // load new sudoku into the comboboxes
                LoadFile(m_openFileDialog1.FileName);
            }
        }

        /// <summary>
        /// This method gets called when the save button was clicked. It
        /// will open a file dialog to enable the user to define the file
        /// to which he wants the current Sudoku to be saved to.
        /// </summary>
        private void SaveButton1_Click(object sender, EventArgs e)
        {
            // open 'save file dialog'
            if (m_saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                // save current sudoku to the choosen file
                SaveFile(m_saveFileDialog1.FileName);
            }
        }

        /// <summary>
        /// This method gets called when the reset button was clicked. It
        /// will reset the comboboxes to their default values (no number
        /// chosen). This method checks first if there are values present,
        /// that were inserted by the solving algorithm. If so, it will only
        /// reset comboboxes with the solution values in it. If only user
        /// inserted numbers are present, this method will set all comboboxes
        /// back to its inital values.
        /// </summary>
        private void ResetButton1_Click(object sender, EventArgs e)
        {
            // first time reset button only erases the green numbers
            // second time all numbers

            // these buttons are no longer valid
            this.PrevButton1.Enabled = false;
            this.NextButton1.Enabled = false;

            // check how many green numbers are set
            int solutionComboboxes = 0;
            for (int j = 0; j < 9; j++)
            {
                for (int i = 0; i < 9; i++)
                {
                    if (this.comboBoxes[j, i].BackColor == m_solutionColor) solutionComboboxes++;
                }
            }

            // if there are solution comboboxes only delete them
            if (solutionComboboxes > 0)
            {
                for (int j = 0; j < 9; j++)
                {
                    for (int i = 0; i < 9; i++)
                    {
                        if (this.comboBoxes[j, i].BackColor == m_solutionColor) this.comboBoxes[j, i].SelectedIndex = -1;
                    }
                }
                return;
            }

            // otherwise reset all comboboxes to no number selected
            for (int j = 0; j < 9; j++)
            {
                for (int i = 0; i < 9; i++)
                {
                    this.comboBoxes[j, i].SelectedIndex = -1;
                }
            }
        }

        /// <summary>
        /// This method gets called when the solve button was clicked. It
        /// will create a new Sudoku instance and fill it with the current
        /// values from the comboboxes.
        /// </summary>
        private void SolveButton1_Click(object sender, EventArgs e)
        {
            // create animation window
            m_animationForm = new SolveAnimationForm1();

            // create Sudoku to solve
            m_currentSudoku = new Sudoku();

            // set user set numbers
            int count = 0;
            for (int j = 0; j < 9; j++)
            {
                for (int i = 0; i < 9; i++)
                {
                    if (this.comboBoxes[j, i].SelectedIndex != -1)
                    {
                        m_currentSudoku.fixNumber(j, i, this.comboBoxes[j, i].SelectedIndex);
                        count++;
                    }
                }
            }

            // check if at least one field is set
            if (count == 0)
            {
                MessageBox.Show("I guess you do not want to caclulate all 6,670,903,752,021,072,936,960 possible Sudokus...   :)", "Too much time?");
                return;
            }

            // start thread to solve sudoku
            this.m_currentSudoku.solve(new Sudoku.resultsFound(this.m_animationForm.closeAnimationWindow), new Sudoku.updateProgress(this.m_animationForm.setProgressBar));

            // show solving animation
            if (m_animationForm.ShowDialog() == DialogResult.Cancel)
            {
                // kill remaining solving threads
                m_currentSudoku.cancelCalculation();
            }
            else
            {
                // check for solutions

                // check for errors
                if (!this.m_currentSudoku.completeSolutionExists())
                {
                    MessageBox.Show("Can't find final solution! Showing partial results.");
                }
                // check how many solutions where found
                else if (this.m_currentSudoku.getNumberOfSolutions() > 1)
                {
                    MessageBox.Show("Sudoku has several Solutions! Showing first solution.");
                    this.PrevButton1.Enabled = true;
                    this.NextButton1.Enabled = true;
                }
                else
                {
                    this.PrevButton1.Enabled = false;
                    this.NextButton1.Enabled = false;
                }

                // show solution
                this.fillComboboxes();
            }

            // dispose animation window
            m_animationForm.Dispose();
        }

        /// <summary>
        /// This method gets called every time the selected number of a combobox
        /// changes. If there is a number selected the background color of the combobox
        /// will be changed.
        /// </summary>
        private void comboBoxes_SelectedIndexChanged(object sender, EventArgs e)
        {
            // change the background color of the combobox
            ComboBox currentComboBox = (ComboBox)sender;
            if (currentComboBox.SelectedIndex != -1) currentComboBox.BackColor = m_fixedColor;
            else currentComboBox.BackColor = m_backgroundColor;
        }

        /// <summary>
        /// This method gets called every time a key is pressed while a combobox has
        /// the focus. It checks for the delete key to be pressed, so that the
        /// user can deselect a number in each combobox. This method is needed to
        /// capture "state" keys like ctrl, alt, del.
        /// </summary>
        private void comboBoxes_KeyDown(object sender, KeyEventArgs e)
        {
            // check for del key down
            ComboBox currentComboBox = (ComboBox)sender;
            if (e.KeyData == Keys.Delete) currentComboBox.SelectedIndex = -1;
        }

        /// <summary>
        /// This method gets called every time a key is pressed while a combobox
        /// has the focus. This method allows the user to press a number key to
        /// select the corresponding number in the combobox.
        /// </summary>
        private void comboBoxes_KeyPress(object sender, KeyPressEventArgs e)
        {
            // check for the buttons 1-9 or backspace
            ComboBox currentComboBox = (ComboBox)sender;
            if ((e.KeyChar >= '1') && (e.KeyChar <= '9'))
            {
                int number = e.KeyChar - '1';
                currentComboBox.SelectedIndex = number;
                return;
            }
            if (e.KeyChar == '\b')
            {
                currentComboBox.SelectedIndex = -1;
                return;
            }
        }
        
    }
}